import argparse
import sys
from dataclasses import asdict, astuple, dataclass
from itertools import product

from tabulate import tabulate
from tqdm import tqdm

import dwdsmor
from dwdsmor.tag import (
    all_tags,
    boundary_tags,
    inflection_tag_seqs,
    info_tags,
    lexeme_tags,
    tag_values,
)
from dwdsmor.traversal import Traversal

headers = {
    "word": "Wordform",
    "spec": "Analysis",
    "analysis": "Lemma",
    "segmented": "Segmented Wordform",
    "lidx": "Lemma Index",
    "pidx": "Paradigm Index",
    "process": "Process",
    "means": "Means",
    "pos": "POS",
    "category": "Subcategory",
    "auxiliary": "Auxiliary",
    "degree": "Degree",
    "person": "Person",
    "gender": "Gender",
    "case": "Case",
    "number": "Number",
    "inflection": "Inflection",
    "nonfinite": "Nonfinite",
    "function": "Function",
    "mood": "Mood",
    "tense": "Tense",
    "metainfo": "Metalinguistic",
    "orthinfo": "Orthography",
    "ellipinfo": "Ellipsis",
    "charinfo": "Characters",
}


def get_taggings(pos, with_info=False):
    tag_seqs = inflection_tag_seqs.get(pos) or inflection_tag_seqs.get("")
    if with_info:
        tag_seqs(ts + info_tags for ts in tag_seqs)

    taggings = set()
    for tag_seq in tag_seqs:
        tagging_values = []
        for tag_type in tag_seq:
            tag_type_values = (
                tag_values.get(pos, dict()).get(tag_type) or tag_values[""][tag_type]
            )
            tagging_values.append(tag_type_values | {None})
        for tagging in product(*tagging_values):
            taggings.add(tuple((t for t in tagging if t is not None)))
    return taggings


@dataclass
class Result(Traversal):
    word: str | None = None

    @staticmethod
    def from_traversal(traversal: Traversal, word):
        return Result(**asdict(traversal), word=word)

    @staticmethod
    def from_spec(spec: str, word: str):
        return Result(**asdict(Traversal.parse(spec)), word=word)


def main():
    arg_parser = argparse.ArgumentParser(description="Traverse DWDSmor automata.")
    arg_parser.add_argument("-g", "--generate", help="Generate", action="store_true")
    arg_parser.add_argument(
        "-s", "--silent", help="Do not report progress", action="store_true"
    )
    arg_parser.add_argument("words", nargs="*")
    args = arg_parser.parse_args()

    results = []
    automata = dwdsmor.automata()

    if args.generate:
        analyzer = automata.analyzer("finite")
        generator = automata.generator("finite")
        for word in args.words:
            lexeme_specs = set()
            for traversal in analyzer.analyze(word):
                traversal = traversal.reparse(visible_boundaries=boundary_tags)
                lexeme_spec = traversal.analysis
                for lexeme_tag in lexeme_tags:
                    tag = getattr(traversal, lexeme_tag)
                    if tag is not None:
                        lexeme_spec += f"<{tag}>"
                lexeme_specs.add((traversal.pos, lexeme_spec))
            for pos, lexeme_spec in sorted(lexeme_specs):
                taggings = get_taggings(pos)
                if not args.silent:
                    taggings = tqdm(
                        taggings,
                        desc=f"Inflection patterns ({pos})",
                        total=len(taggings),
                    )
                for tagging in taggings:
                    spec = lexeme_spec + "".join(f"<{tag}>" for tag in tagging)
                    for generated in generator.generate(spec):
                        results.append(Result.from_spec(spec, generated.spec))
    else:
        analyzer = automata.analyzer("lemma")
        for word in args.words:
            max_boundaries = {}
            for traversal in analyzer.analyze(word):
                with_boundaries = traversal.reparse(
                    visible_boundaries=boundary_tags, boundary_tag="|"
                )
                boundaries = with_boundaries.analysis.count("|")
                # group by traversal props, leaving out discriminating
                # spec, which expresses boundaries
                group_by = astuple(traversal)[1:]
                if max_boundaries.get(group_by, -1) < boundaries:
                    max_boundaries[group_by] = boundaries
                    results.append(Result.from_traversal(traversal, word))

    if not results:
        return 1

    results = sorted(results, key=lambda r: r.spec)

    def has_value(tag_type, results=results):
        for result in results:
            if getattr(result, tag_type) is not None:
                return True
        return False

    cols = [
        "word",
        "analysis",
        "spec",
        *(tt for tt in all_tags if has_value(tt)),
    ]

    print(
        tabulate(
            [[getattr(r, c) for c in cols] for r in results],
            headers=[headers.get(c, c) for c in cols],
            tablefmt="github",
        )
    )

    return 0


if __name__ == "__main__":
    sys.exit(main())



"""
Handling of SFST automata/transducer collections on which DWDSmor is based.

For different use cases like lemmatization, paradigm generation or
morphological segmentation, DWDSmor relies on different transducers
based on a common lexicon. Those transducers are stored collectively
in a common local directory from which they can be read and used via
the class ``Automata`` in this module.
"""

__all__ = [
    "detect_dev_root_dir",
    "detect_root_dir",
    "Generator",
    "Analyzer",
    "Automata",
    "automata",
    "Lemmatizer",
    "lemmatizer",
    "load_from_hub",
]

import csv
import lzma
import os
from collections import defaultdict
from dataclasses import dataclass
from os import PathLike
from pathlib import Path
from typing import Dict, Iterable, List, Optional, Union

from dotenv import dotenv_values
from huggingface_hub import snapshot_download
from huggingface_hub.utils import disable_progress_bars, enable_progress_bars
from sfst_transduce import CompactTransducer, Transducer

from .log import logger
from .traversal import Traversal
from .version import __version__

config = {
    **dotenv_values(".env.shared"),
    **dotenv_values(".env"),
    **os.environ,
}


class AutomataDirNotFound(Exception):
    pass


def detect_dev_root_dir():
    """
    Detect development environment and return one of its automaton
    dirs.

    Development environments are flagged via a set environment
    variable ``DWDSMOR_ENV`` (i. e. in ``.env.shared``). In such
    environments, one of the automata in a subdirectory of ``build/``,
    i.e. ``build/open`` is returned.
    """
    if config.get("DWDSMOR_DEV"):
        build_dir = Path("build")
        for edition in ("dwds", "open", "dev"):
            edition_dir = build_dir / edition
            if edition_dir.is_dir():
                return str(edition_dir)
        raise AutomataDirNotFound(
            "$DWDSMOR_DEV is set, but no known automata available under build/"
        )


def detect_root_dir() -> Path:
    """
    Detect local automata directory if none is specified.

    1. Try to get it from an environment variable ``DWDSMOR_AUTOMATA_DIR``.
    2. Try to detect a development environment and get it from there.
    3. Raise ``AutomataDirNotFound``.
    """
    root = config.get("DWDSMOR_AUTOMATA_DIR")
    if root:
        return Path(root)
    root = detect_dev_root_dir()
    if root:
        return Path(root)
    raise AutomataDirNotFound()


@dataclass
class Generator:
    """
    Wrap a SFST transducer for generating strings, i.e. inflection paradigms.
    """

    transducer: Transducer

    def generate(self, s: str) -> Iterable[Traversal]:
        """
        Generate all traversals of a transducer for the given input string.
        """
        return (Traversal.parse(traversal) for traversal in self.transducer.generate(s))


@dataclass
class Analyzer:
    """
    Wrap a SFST transducer providing a morphological analysis of input strings.
    """

    transducer: CompactTransducer

    def analyze(self, s: str) -> Iterable[Traversal]:
        """
        Generate all traversals of a transducer that express
        morphological analyses of an input string.
        """
        return (Traversal.parse(traversal) for traversal in self.transducer.analyse(s))


automaton_types = {"finite", "index", "lemma", "morph", "root"}
traversal_automaton_types = {"index"}


def assert_valid_automaton_type(automaton_type: str, types=automaton_types):
    assert automaton_type in automaton_types, (
        f"{automaton_type} is not a valid automaton_type. "
        f"Supported types: {automaton_types}"
    )


class Automata:
    def __init__(self, root_dir: Optional[Union[str, PathLike]] = None):
        self.root_dir = Path(root_dir) if root_dir else detect_root_dir()

    def transducer(
        self, automaton_type="lemma", generate=True, both_layers=False
    ) -> Union[Transducer, CompactTransducer]:
        file_suffix = "a" if generate else "ca"
        transducer_file = Path(self.root_dir) / f"{automaton_type}.{file_suffix}"
        assert transducer_file.is_file(), f"{transducer_file} does not exist"
        if not generate:
            transducer = CompactTransducer(transducer_file.as_posix())
            transducer.both_layers = both_layers
            return transducer
        else:
            return Transducer(transducer_file.as_posix())

    def generator(self, automaton_type="index") -> Generator:
        assert_valid_automaton_type(automaton_type)
        return Generator(self.transducer(automaton_type))  # type: ignore

    def analyzer(self, automaton_type="lemma", both_layers=False) -> Analyzer:
        assert_valid_automaton_type(automaton_type)
        return Analyzer(
            self.transducer(
                automaton_type, generate=False, both_layers=both_layers
            )  # type: ignore
        )

    def traversals(self, automaton_type="index") -> Dict[str, List[Traversal]]:
        assert_valid_automaton_type(automaton_type, types=traversal_automaton_types)
        traversals_file = Path(self.root_dir) / f"{automaton_type}.csv.lzma"
        traversals = defaultdict(list)
        with lzma.open(traversals_file, "rt") as traversals_csv:
            for traversal in csv.DictReader(traversals_csv):
                traversal_obj = Traversal(**traversal)
                traversals[traversal_obj.analysis].append(traversal_obj)
        return traversals


default_repo_id = config.get("DWDSMOR_HF_REPO_ID", "zentrum-lexikographie/dwdsmor-open")


def load_from_hub(
    repo_id: Optional[str] = None, revision: Optional[str] = None, **kwargs
):
    repo_id = repo_id or default_repo_id
    revision = revision or f"v{__version__}"
    logger.debug("Load automata from Huggingface repo %s @ %s", repo_id, revision)
    try:
        disable_progress_bars()
        return Automata(snapshot_download(repo_id=repo_id, revision=revision, **kwargs))
    finally:
        enable_progress_bars()


def automata(automata_location: Optional[str] = None, *args, **kwargs):
    if automata_location is not None:
        path = Path(automata_location)
        if path.is_dir():
            logger.debug("Load automata from local dir '%s'", str(path))
            return Automata(path)
    if automata_location is None:
        try:
            detected_dir = detect_root_dir()
            logger.debug(
                "Load automata from detected local dir '%s'", str(detected_dir)
            )
            return Automata(detected_dir)
        except AutomataDirNotFound:
            pass
    return load_from_hub(automata_location, *args, **kwargs)


class Lemmatizer:
    def __init__(self, automata, automaton_type="lemma"):
        self.analyzer = automata.analyzer(automaton_type)

    def __call__(self, word, **criteria):
        traversals = tuple(self.analyzer.analyze(word))
        criteria_stack = list(criteria.items())
        criteria_stack.reverse()
        while criteria_stack:
            if len(traversals) == 1:
                (traversal,) = traversals
                return traversal.analysis
            attr, attr_vals = criteria_stack.pop()
            filtered = tuple((t for t in traversals if getattr(t, attr) in attr_vals))
            traversals = filtered or traversals
        for traversal in traversals:
            return traversal.analysis


def lemmatizer(*args, automaton_type="lemma", **kwargs):
    return Lemmatizer(automata(*args, **kwargs), automaton_type)



lexeme_tags = ("lidx", "pidx", "pos")
inflection_tags = (
    "category",
    "degree",
    "function",
    "person",
    "gender",
    "case",
    "number",
    "nonfinite",
    "tense",
    "mood",
    "auxiliary",
    "inflection",
)
info_tags = ("metainfo", "orthinfo", "charinfo", "ellipinfo")
wordformation_tags = ("processes", "means")

all_tags = (*lexeme_tags, *inflection_tags, *info_tags, *wordformation_tags)

tag_values = {
    "": {
        "lidx": {f"IDX{i}" for i in range(1, 9)},
        "pidx": {f"PAR{i}" for i in range(1, 9)},
        "pos": {
            "+ADJ",
            "+ADV",
            "+ART",
            "+CARD",
            "+CONJ",
            "+DEM",
            "+FRAC",
            "+INDEF",
            "+INTJ",
            "+NN",
            "+NPROP",
            "+ORD",
            "+POSS",
            "+POSTP",
            "+PPRO",
            "+PREP",
            "+PREPART",
            "+PROADV",
            "+PTCL",
            "+PUNCT",
            "+REL",
            "+V",
            "+WPRO",
        },
        "degree": {"Pos", "Comp", "Sup"},
        "function": {"Attr", "Subst", "UnmFunc"},
        "person": {"1", "2", "3"},
        "gender": {"Fem", "Neut", "Masc", "UnmGend"},
        "case": {"Nom", "Gen", "Dat", "Acc", "UnmCase"},
        "number": {"Sg", "Pl", "UnmNum"},
        "nonfinite": {"Inf", "Part"},
        "tense": {"Pres", "Past", "Perf"},
        "mood": {"Ind", "Subj", "Imp"},
        "auxiliary": {"haben", "sein"},
        "inflection": {"St", "Wk", "UnmInfl"},
        "metainfo": {"Old", "NonSt"},
        "orthinfo": {"OLDORTH", "CH"},
        "charinfo": {"CAP"},
        "ellipinfo": {"TRUNC"},
        "processes": {"COMP", "DER", "CONV"},
        "means": {"concat", "hyph", "ident", "pref", "prev", "suff"},
    },
    "+ADJ": {"function": {"Attr", "Subst", "Attr/Subst", "Pred/Adv"}},
    "+ART": {
        "category": {"Def", "Indef", "Neg"},
    },
    "+CONJ": {"category": {"AdjComp", "Coord", "Sub", "InfCl"}},
    "+ORD": {
        "function": {"Attr/Subst"},
    },
    "+POSS": {
        "function": {"Attr/Subst"},
    },
    "+PPRO": {
        "category": {"Pers", "Refl", "Rec"},
    },
    "+PTCL": {"category": {"AdjPos", "AdjSup", "InfCl", "Neg"}},
    "+PUNCT": {
        "category": {
            "Comma",
            "Period",
            "Ellip",
            "Quote",
            "Paren",
            "Dash",
            "Slash",
            "Other",
        },
    },
    "+V": {"category": {"NonCl", "Cl"}},
}


tag_types = {
    v: tt for _pos, tvs in tag_values.items() for tt, vs in tvs.items() for v in vs
}


inflection_boundary_tags = "~"
word_formation_tags = "#-|"
other_boundary_tags = "="
boundary_tags = inflection_boundary_tags + word_formation_tags + other_boundary_tags

inflection_tag_seqs = {
    "": (("function", "gender", "case", "number", "inflection"),),
    "+ADJ": (("degree", "function", "gender", "case", "number", "inflection"),),
    "+ADV": (("degree"),),
    "+ART": (("category", "function", "gender", "case", "number", "inflection"),),
    "+CONJ": (("category",),),
    "+FRAC": (("function",),),
    "+INTJ": ((),),
    "+NN": (("gender", "case", "number", "inflection"),),
    "+NPROP": (("gender", "case", "number", "inflection"),),
    "+PPRO": (("category", "person", "gender", "case", "number"),),
    "+PREP": ((),),
    "+PREPART": (("gender", "case", "number"),),
    "+PROADV": ((),),
    "+PTCL": (("category"),),
    "+V": (
        ("mood", "number"),
        ("nonfinite", "category"),
        ("nonfinite", "tense", "auxiliary"),
        ("person", "number", "tense", "mood"),
    ),
}



from dataclasses import dataclass

from dwdsmor.tag import boundary_tags, tag_types


@dataclass
class Traversal:
    spec: str
    surface: str
    analysis: str
    lidx: str | None = None
    pidx: str | None = None
    pos: str | None = None
    category: str | None = None
    degree: str | None = None
    function: str | None = None
    person: str | None = None
    gender: str | None = None
    case: str | None = None
    number: str | None = None
    nonfinite: str | None = None
    tense: str | None = None
    mood: str | None = None
    auxiliary: str | None = None
    inflection: str | None = None
    metainfo: str | None = None
    orthinfo: str | None = None
    charinfo: str | None = None
    ellipinfo: str | None = None
    processes: str | None = None
    means: str | None = None

    def reparse(self, **args):
        return Traversal.parse(self.spec, **args)

    @staticmethod
    def parse_label(s, visible_boundaries, boundary_tag):
        c = s[0]
        if c != "<" and c != ":":
            return "", c, s[1:]
        elif c == "<":
            e = s.find(">")
            assert e > 0
            tag = s[1:e]
            s = s[e + 1 :]
            if tag == "":
                return "", "", s
            elif tag in visible_boundaries:
                return ("", boundary_tag or f"<{tag}>", s)
            elif tag in boundary_tags:
                return ("", "", s)
            else:
                return tag_types[tag], tag, s

    @staticmethod
    def parse(s, visible_boundaries="", boundary_tag=None):
        spec = s
        surface = ""
        analysis = ""
        tags = {}
        while s:
            if s.startswith(r"\:"):
                surface += ":"
                analysis += ":"
                s = s[2:]
                continue
            k, v, s = Traversal.parse_label(s, visible_boundaries, boundary_tag)
            if k:
                tags[k] = v
            else:
                analysis += v
            if s.startswith(":"):
                k, v, s = Traversal.parse_label(s[1:], visible_boundaries, boundary_tag)
                if k:
                    tags[k] = v
                else:
                    surface += v
            elif not k:
                surface += v
        return Traversal(spec, surface, analysis, **tags)



import logging
import os
import sys

logging.basicConfig(
    format="%(asctime)s - %(name)s - %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
    stream=sys.stderr,
    level=(logging.DEBUG if os.getenv("DEBUG") else logging.INFO),
)

logger = logging.getLogger("dwdsmor")

__all__ = ["logger"]
